
#define FUSE_USE_VERSION 22
#include <fuse/fuse.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>
#include <malloc.h>

static int force=0;
static int verbose=0; 
static int read_only=0; 

typedef struct partition_t partition_t;

struct partition_t {
  int id;
  char type;
  off_t offset;
  off_t size;
  partition_t *next;
};


partition_t* parse_partition(char* buf){
  if (!buf[4])
    return NULL;
  partition_t* j = malloc(sizeof(partition_t));
  j->type = buf[4];
  j->offset = ((off_t)((buf[0x08]  & 0x000000ff) 
                       | (buf[0x09] <<  8 & 0x0000ff00) 
                       | (buf[0x0a] << 16 & 0x00ff0000)
                       | (buf[0x0b] << 24 & 0xff000000)) << 9);
  j->size   = ((off_t)((buf[0x0c]  & 0x000000ff) 
                       | (buf[0x0d] <<  8 & 0x0000ff00) 
                       | (buf[0x0e] << 16 & 0x00ff0000)
                       | (buf[0x0f] << 24 & 0xff000000)) << 9);

  if (verbose){
    if (j->type == (char)0x05 || j->type == (char)0x85 || j->type == (char)0x0f){
      char* type = "unknown";

      switch (j->type){
      case (char)0x05:
        type = "MS-DOS";
        break;
      case (char)0x85:
        type = "Linux";
        break;
      case (char)0x0f:
        type = "Windows";
        break;
      }

      printf("Extended partition at offset: %016llx size: %016llx (%s)\n", 
             j->offset, j->size, type);
    }else{
      printf("Partition type: %02hhx at offset: %016llx size: %016llx\n"
             , j->type, j->offset, j->size);  
    }
  }

  return j;
}

static void is_mbr(int fd, off_t offset){
  char buf[2];
  lseek(fd, offset+0x1fe, SEEK_SET);
  read(fd, buf, 2);
  if (!(buf[0]==(char)0x55 && buf[1]==(char)0xaa)){
    fprintf(stderr,"Invalid magic at 0x%016llx: is 0x%02hhx 0x%02hhx should be 0x55 0xaa\n",offset+0x1fe, buf[0], buf[1]);
    if (!force){
      fputs("File isn't valid PC disk image or partition tale is damaged.\n",stderr);
      fputs("Use -f to override this check.\n",stderr);
      exit(4);
    }
  }
}

static partition_t* read_partition_table(int fd){
  partition_t* i = NULL;
  partition_t* ext = NULL;
  char buf[16];
  int counter = 0;

  is_mbr(fd,0);
  lseek(fd, 0x1be, SEEK_SET);
  for(counter=1; counter < 5; counter++){
    read(fd, buf, 16);
    partition_t *j = parse_partition(buf);
    if (j){
      j->next = i;
      j->id = counter;
      if (j->type == (char)0x05 || j->type == (char)0x85 || j->type == (char)0x0f){
        assert(!ext);
        ext = j;
      }
      i = j;
    }
  }
  
  counter = 4;

  partition_t *e = ext;

  while (e){
    is_mbr(fd,e->offset);
    partition_t *cur = e;
    int it;
    e = NULL;
    lseek(fd, cur->offset+0x1be, SEEK_SET);
    for (it=0; it<4; ++it){
      read(fd, buf, 16);
      partition_t *j = parse_partition(buf);
      if (j){
        j->offset += ext->offset;
        if (j->type == (char)0x05 || j->type == (char)0x85 || j->type == (char)0x0f){
          e = j;
        }else{
          j->next = i;
          j->id = ++counter;
          i = j;          
        }
      }
    }
    if (cur!=ext)
      free(cur);
  }

  return i;
}

static partition_t* part = NULL;
static partition_t** ptable = NULL;
static size_t ptable_len = 0;
static int fd;
static char* fname;
static struct stat def_stat;

static void create_ptable(partition_t* list){
  int max_id=0;
  partition_t *i = list;
  part = list;

  while(i){
    if(i->id > max_id){
      max_id = i->id;
    }
    i=i->next;
  }
  
  ptable_len = max_id;
  ptable = calloc(max_id+1,sizeof(partition_t*));
  assert(ptable);

  i = list;
  while(i){
    ptable[i->id] = i;
    i=i->next;
  }

}

static partition_t* get_entry(int id){
  if (id > ptable_len){
    if (verbose)
      fprintf(stderr,"Invalid partition number: %d, maximum is %d\n", id, ptable_len);
    return NULL;
  }

  return ptable[id];
}

static int pfs_getattr(const char *path, struct stat *stbuf)
{

  memcpy(stbuf, &def_stat, sizeof(struct stat));

  if (strcmp(path,"/")==0){
      stbuf->st_mode &= ~S_IFMT;
      stbuf->st_mode |= S_IFDIR;
      return 0;    
  }

  if (path[0]!='/')
    return -ENOENT;

  int id = (int)atoi(path+1);
  partition_t *i = get_entry(id);

  if(i){
    stbuf->st_mode &= ~S_IFMT;
    stbuf->st_mode |= S_IFREG;
    stbuf->st_size = i->size;
    return 0;
  }

  return -ENOENT;
}


static int pfs_getdir(const char *path, fuse_dirh_t h, fuse_dirfil_t filler)
{
  partition_t *i = part;


  if (strcmp(path,"/")!=0)
    return -ENOENT;

  while(i){
    char buf[16];
    snprintf(buf, 16, "%d",i->id);
    filler(h,buf,0,0);
    i=i->next;
  }

  return 0;
}



static int pfs_open(const char *path, struct fuse_file_info *fi)
{
    int res;

    int id = (int)atoi(path+1);
 
    partition_t *i = get_entry(id);
    
    if(i){
      fi->fh = id;   
      return 0;
    }

    return -ENOENT;
}

static int pfs_read(const char *path, char *buf, size_t size, off_t offset,
                    struct fuse_file_info *fi)
{
  partition_t *i = get_entry(fi->fh);
  size_t asize = size;
  int fd, res;
  
  if (!i)
    return -EBADF;

  if (i->size < offset)
    return 0;
  if (i->size < (offset+size))
    asize -= (offset+size) - i->size;

  fd = open(fname, O_RDONLY); /* We need separate fd for each thread to avoid 
                               * race condition in time between lseek and 
                               * read/write
                               *
                               * This is trivial way how to accomplish this
                               * requirement.
                               */
  if (fd == -1){
    int err=errno;
    if (verbose)
      perror("opening image for reading");
    return -err;
  }

  lseek(fd, i->offset+offset, SEEK_SET);
  res = read(fd, buf, asize);
  close(fd);

  if (res ==-1)
    return -errno;
  return res;
}

static int pfs_write(const char *path, const char *buf, size_t size,
                     off_t offset, struct fuse_file_info *fi)
{
  partition_t *i = get_entry(fi->fh);
  size_t asize = size;
  int fd, res;

  if (read_only)
    return -EROFS;

  if (!i)
    return -EBADF;

  if (i->size < offset)
    return -ENOSPC;
  if (i->size < (offset+size))
    asize -= (offset+size) - i->size;

  fd = open(fname, O_WRONLY); /* We need separate fd for each thread to avoid 
                               * race condition in time between lseek and 
                               * read/write
                               *
                               * This is trivial way how to accomplish this
                               * requirement.
                               */
  if (fd == -1){
    int err=errno;
    if (verbose)
      perror("opening image for writing");
    return -err;
  }

  lseek(fd, i->offset+offset, SEEK_SET);
  res = write(fd, buf, asize);
  close(fd);

  if (res ==-1)
    return -errno;
  return res;
}



static struct fuse_operations pfs_oper = {
    .getattr	= pfs_getattr,
    .getdir	= pfs_getdir,
    .open	= pfs_open,
    .read	= pfs_read,
    .write	= pfs_write,
};

static void usage(){
  fputs("usage: pfs [-dfrv] <image> <mount-point>\n",stderr);
}

static void help(){
  usage();
  fputs("\nOptions:\n",stderr);
  fputs(  "        -d  Daemonize after startup\n",stderr);
  fputs(  "        -f  Ignore partition table magic value\n",stderr);
  fputs(  "        -r  Read only acces\n",stderr);
  fputs(  "        -v  Print more information\n",stderr);
  
}

int main(int argc, char *argv[])
{
  int fuse_channel;
  int fd;
  int c;
  int daemon_flag=0;
  struct fuse *fuse;
  char *opts;
  char *mount;
  
  while ((c = getopt (argc, argv, "fvrdh")) != -1)
    switch (c)
      {
      case 'f':
        force = 1;
        break;
      case 'v':
        verbose = 1;
        break;
      case 'd':
        daemon_flag = 1;
        break;
      case 'r':
        read_only = 1;
        break;
      case 'h':
        help();
        return 0;
        break;
      case '?':
        if (isprint (optopt))
          fprintf (stderr, "Unknown option `-%c'.\n", optopt);
        else
          fprintf (stderr,
                   "Unknown option character `\\x%x'.\n",
                   optopt);
        usage();
        return 1;
      default:
        abort ();
      }

  if (argc != optind+2){
    usage();
    return 1;
  }
  
  mount = argv[optind+1];
  fname = argv[optind];

  fd = open(fname, O_RDONLY);
  if (fd==-1){
    perror(fname);
    return 3;
  }
  fstat(fd, &def_stat);
  create_ptable(read_partition_table(fd));
  close(fd);

  opts = malloc(strlen(fname)+10);

  strcpy(opts, "fsname=");
  strcpy(opts+7, fname);

  fuse_channel = fuse_mount(mount,opts);

  if (fuse_channel == -1){
    fputs("Cannot mount filesystem.\n",stderr);
    return 2;
  }

  fuse = fuse_new(fuse_channel, NULL, &pfs_oper, sizeof(pfs_oper));

  fflush(stdout);
  fflush(stderr);

  if (daemon_flag)
    daemon(1,1);

  fuse_loop(fuse);

  fuse_unmount(mount);
}
